/*
 * 1. Open up the project properties
 * 2. Visit C/C++ General > Preprocessor Include Paths, Macros, etc
 * 3. Select the Providers tab
 * 4. Check the box for "CDT GCC Built-in Compiler Settings"
 * 5. Set the compiler spec command to "xtensa-esp32-elf-gcc ${FLAGS} -E -P -v -dD "${INPUTS}""
 * 6. Rebuild the index
 */

#include <stdio.h>
#include <string.h>

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/timers.h"

#include <esp_system.h>
#include <nvs_flash.h>
#include <sys/param.h>

#include "esp_err.h"
#include "nvs_flash.h"

//app driver includes
#include <mbedtls/pk.h>
#include <mbedtls/error.h>
#include <mbedtls/entropy.h>
#include <mbedtls/ctr_drbg.h>
#include "mbedtls/bignum.h"
#include "mbedtls/x509.h"
#include "mbedtls/rsa.h"
#include "mbedtls/config.h"
#include "mbedtls/platform.h"
#include "mbedtls/base64.h"
#include "esp_partition.h"

static unsigned char privKeyPem[2000];
static unsigned char pubKeyPem[1000];

uint32_t privKeyOffset = 0 * CONFIG_WL_SECTOR_SIZE;
uint32_t pubKeyOffset = 1 * CONFIG_WL_SECTOR_SIZE;
uint32_t tamperDataOffset = 2 * CONFIG_WL_SECTOR_SIZE;
esp_partition_t *partStorage = NULL;

mbedtls_pk_context pk;
mbedtls_entropy_context entropy;
mbedtls_ctr_drbg_context ctrDrbg;

const char personalizationString[11] = "rsa_genkey";	//requires static size allocation in classes
#define KEY_SIZE 2048
#define EXPONENT 65537

bool Crypto_Verify(bool initisalised);
bool Crypto_LoadKeys();
bool Crypto_GenerateKeyPair(void);

//uint8_t test[4];

extern "C" void app_main() {
	//Assign partition regions
	partStorage = (esp_partition_t *)esp_partition_find_first(ESP_PARTITION_TYPE_DATA, ESP_PARTITION_SUBTYPE_ANY, "storage");

	if (partStorage != NULL) {
		if (!Crypto_LoadKeys()) {
			if (!Crypto_GenerateKeyPair()) {
				puts("!ERROR! Unable to create key pair!");
			}
		}

		//allocate inside function
		uint8_t test[4];

		//test bad alignment...maybe?
		for (int i = 0; i < sizeof(test); i++) {
			esp_partition_read(partStorage, tamperDataOffset, &test[i], 1);
		}

		printf("Test read %d\n", test[0]);
	}
	else {
		puts("Couldn't find partition");
	}
}

bool Crypto_LoadKeys() {
	bool ret = true;
	esp_err_t resp = esp_partition_read(partStorage, privKeyOffset, privKeyPem, sizeof(privKeyPem) - 1);
	if (ESP_OK != resp) {
		puts("Err: Couldn't load priv key from partition");
		return false;
	}

	resp = esp_partition_read(partStorage, pubKeyOffset, pubKeyPem, sizeof(pubKeyPem) - 1);
	if (ESP_OK != resp) {
		puts("Err: Couldn't load pub key from partition");
		ret = false;
	}
	else if (NULL == strstr((const char *)privKeyPem, "-----BEGIN RSA PRIVATE KEY-----") || NULL == strstr((const char *)privKeyPem, "-----END RSA PRIVATE KEY-----")) {
		puts("Err: Priv key couldn't be loaded");
		ret = false;
	}
	else if (NULL == strstr((const char *)pubKeyPem, "-----BEGIN PUBLIC KEY-----") || NULL == strstr((const char *)pubKeyPem, "-----END PUBLIC KEY-----")) {
		puts("Err: Pub key couldn't be loaded");
		ret = false;
	}

	/* pk struct not currently initialised */
	if (!Crypto_Verify(false)) {
		puts("Err: Private key couldn't verified");
		ret = false;
	}

	//null terminate safety
	privKeyPem[sizeof(privKeyPem) - 1] = 0;
	pubKeyPem[sizeof(pubKeyPem) - 1] = 0;

	return ret;
}

bool Crypto_Verify(bool initisalised) {
	int ret = 0;
	/* Key has no password - parse in NULL size 0 */
	if (!initisalised) {
		mbedtls_pk_init(&pk);
		if ((ret = mbedtls_pk_parse_key(&pk, privKeyPem, strlen((const char *)privKeyPem) + 1, NULL, 0)) != 0) {
			printf("Unable to parse %d\n", ret);
			goto exit;
		}
	}

	/* Asumme keys are already loaded into internal variables */
	if ((ret = mbedtls_rsa_check_privkey(mbedtls_pk_rsa(pk))) != 0) {
		puts("Err: RSA context does not contain an rsa private key");
		goto exit;
	}

	if ((ret = mbedtls_rsa_check_pubkey(mbedtls_pk_rsa(pk))) != 0) {
		puts("Err: RSA context does not contain an rsa public key");
		goto exit;
	}

	puts("RSA keys confirmed");

exit:
	if (!initisalised) {
		mbedtls_pk_free(&pk);
	}
	return !ret;
}

bool Crypto_GenerateKeyPair(void) {
	bool ret = false;	 //== ESP_OK

	/* Initialise variables */
	mbedtls_entropy_init(&entropy);
	mbedtls_ctr_drbg_init(&ctrDrbg);
	mbedtls_pk_init(&pk);

	puts("Seeding the random number generator...");
	if ((ret = mbedtls_ctr_drbg_seed(&ctrDrbg, mbedtls_entropy_func, &entropy, (const unsigned char *)personalizationString, strlen(personalizationString))) != 0) {
		printf("Failed to seed rng\nmbedtls_ctr_drbg_seed returned %d\n", ret);
		goto cleanup;
	}

	if ((ret = mbedtls_pk_setup(&pk, mbedtls_pk_info_from_type(MBEDTLS_PK_RSA))) != 0) {
		printf("pk_setup failed: %i\n", ret);
	}

	printf(" ok\n  . Generating the RSA key [ %d-bit ]...", KEY_SIZE);

	/* force task yield before generate a key that could take several seconds to do */
	taskYIELD();
	puts("Generating a key pair takes up to 20s and will likely trigger task watchdog during this process");

	if ((ret = mbedtls_rsa_gen_key(mbedtls_pk_rsa(pk), mbedtls_ctr_drbg_random, &ctrDrbg, KEY_SIZE, EXPONENT)) != 0) {
		printf(" failed\n  ! mbedtls_rsa_gen_key returned %d\n\n", ret);
		goto cleanup;
	}

	/* pk struct already intitialised */
	if (!Crypto_Verify(true)) {
		goto cleanup;
	}

	puts("Writing public key to string(PEM format)....");

	//unsigned char pubKeyPem[1000];
	memset(pubKeyPem, 0, sizeof(pubKeyPem));
	if (mbedtls_pk_write_pubkey_pem(&pk, pubKeyPem, sizeof(pubKeyPem)) != 0) {
		puts("Err: Write public key to string failed");
		goto cleanup;
	}

	puts("Writing private key to string(PEM format)....");

	memset(privKeyPem, 0, sizeof(privKeyPem));
	ret = mbedtls_pk_write_key_pem(&pk, privKeyPem, sizeof(privKeyPem));
	if (ret != 0) {
		printf("write private key to string failed with code %04x\n", ret);
	}

	puts(" ok\n  .");
	puts("Success, Key pair created.");

cleanup:
	mbedtls_pk_free(&pk);
	mbedtls_ctr_drbg_free(&ctrDrbg);
	mbedtls_entropy_free(&entropy);

	return !ret;
}
